using System.Collections.Generic;
using System.Linq;
using System;
using REST_JSON_API;
using Unity.Mathematics;
using UnityEngine;
using System.Runtime.CompilerServices;
using Newtonsoft.Json;

/// <summary>
/// 
/// </summary>
/// <seealso cref="SOMDoc.MakeDeepList(SOMDoc[], SOMDoc)"/>
/// <remarks>Needed to refere to lists serverside</remarks>
public class ListFact : FactWrappedCRTP<ListFact>
{
    public string[] lids = new string[0];

    public SOMDoc[] payload = new SOMDoc[0];

    public OMA ListType;

    public ListFact() : base() { }

    public ListFact(string[] lids, SOMDoc[] payload, SOMDoc ElementType, FactRecorder organizer) : base(organizer)
    {
        Init(lids, payload, ElementType);
        SendToMMT();
    }

    public ListFact(string[] lids, SOMDoc[] payload, SOMDoc ElementType, SOMDoc _ServerDefinition, FactRecorder organizer) : base(organizer)
    {
        Init(lids, payload, ElementType);
        this.ServerDefinition = _ServerDefinition;
    }

    private void Init(string[] lids, SOMDoc[] payload, SOMDoc ElementType)
    {
        this.lids = lids ?? new string[0];

        this.payload = new SOMDoc[math.max(this.lids.Length, payload?.Length ?? 0)];
        payload?.CopyTo(this.payload, 0);

        if (!HasDependentFacts)
        {
            if (payload.Any(p => p == null))
                throw new ArgumentException(nameof(payload) + "must not include null elements iff " + nameof(lids) + " is all nulls or empty!");
        }
        else
        {
            for (int i = 0; i < this.lids.Length; i++)
                if (this.lids[i] != null)
                    this.payload[i] = new OMS(this.lids[i]);
                else
                if (this.payload[i] == null)
                    throw new ArgumentException(nameof(lids) + " and " + nameof(payload) + " have to be complementary not null!");
        }

        if (ElementType == null)
        {
            if (this.payload.Length == 0)
                throw new ArgumentException(nameof(ElementType) + " must not be null if ListFact is empty List!");

            ElementType = payload[0].SOMDocType();
        }

        this.ListType = new OMA(
            new OMS(MMTConstants.ListType),
            new[] { ElementType }
        );
    }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => new ListFact(
            lids.Select(id => id == null ? null : old_to_new[id]).ToArray(),
            payload.Select(somdoc => somdoc?.MapURIs(old_to_new)).ToArray(),
            ListType.arguments[0],
            organizer);

    public static new List<Fact> parseFact(MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol)
            return null;

        if (MMTSymbol.type is not OMA listOMA
         || listOMA.arguments[0] is not OMA typeOMA)
            // try to salvage
            return DynamicListFact.MMTFactory(null,
                MMTSymbol.defines, MMTSymbol.type,
                StageStatic.stage.factState, MMTSymbol.@ref
            );

        List<SOMDoc> payload;

        if (MMTSymbol.defines is OMA OMAList
         && OMAList.applicant is OMS OMSList
         && OMSList.uri == MMTConstants.ListApplicant)
        {
            payload = OMAList.arguments.ToList();
        }
        else
        {
            payload = new();
            SOMDoc next_element = MMTSymbol.defines;
            int depth = 0;
            while (true)
            {
                ++depth;

                if (next_element is not OMA current_element
                 || current_element.applicant is not OMS OMSapp)
                { // neither ListLiteral nor ListEnd
                    throw new ArgumentException("Malformed List-Format");
                }

                if (OMSapp.uri != MMTConstants.ListLiteral)
                {
                    if (OMSapp.uri != MMTConstants.ListEnd)
                    { // neither ListLiteral nor ListEnd
                        throw new ArgumentException("Malformed List-Format");
                    }

                    break;
                }

                if (current_element.arguments.Length < 2) // Maybe empty list?
                    throw new ArgumentException("Unrecognized List-Format");

                payload.Add(current_element.arguments[1]);
                next_element = current_element.arguments[0];
            }
        }

        return new() { new ListFact(
            payload.Select(p => (p as OMS)?.uri).ToArray(),
            payload.ToArray(),
            listOMA.arguments[0],
            fact.@ref, StageStatic.stage.factState) };
    }

    public override MMTFact MakeMMTDeclaration()
        => new MMTGeneralFact(Label, ListType, SOMDoc.MakeShallowList(payload));

    protected override string[] GetDependentFactIds()
        => lids.Where(lid => lid != null).ToArray();

    protected override void RecalculateTransform() { }

    protected override bool EquivalentWrapped(ListFact f1, ListFact f2)
        => f1.ListType.Equivalent(f2.ListType)
        && f1.payload.Length == f2.payload.Length
        && DependentFactsEquivalent(f1, f2)
        && f1.payload
            .Zip(f2.payload, (p1, p2) => SOMDoc.Equivalent(p1, p2))
            .All(b => b);
}

public class DynamicListFact : FactWrappedCRTP<DynamicListFact>, IUnpackable
{
    public List<object> payload;

    public SOMDoc indirect_payload;

    public SOMDoc ListType;

    public DynamicListFact() : base() { }

    public DynamicListFact(FactRecorder organizer) : base(organizer) { }

    public DynamicListFact(List<object> payload, SOMDoc indirect_payload, SOMDoc ListType, FactRecorder organizer) : base(organizer)
    {
        this.payload = payload;
        this.indirect_payload = indirect_payload;
        this.ListType = ListType;

        SendToMMT();
    }

    public static List<Fact> MMTFactory(List<dynamic> payload, SOMDoc indirect_payload, SOMDoc ListType, FactRecorder organizer, SOMDoc _ServerDefinition = null)
    {
        if (payload == null
         && indirect_payload == null)
            return new();

        if (payload == null)
        {
            Func<object[], object[]> builder =
                indirect_payload.PartialInvokeCastingLambdaExpression(out _, out Type[] signature);

            if (signature.Length != 1)
                return new();

            object result = builder(new object[0])[0];

            if (result is IEnumerable<dynamic> enum_result)
                payload = enum_result.ToList();
            else
            if (result is IEnumerable<float> enum_float) // special case for boxed/struct values
                payload = enum_float.Select(f => (dynamic)f).ToList();
            else
            if (signature[0].IsArray)
                throw new NotSupportedException();
            else
                return new();
        }

        //TODO? indirect_payload ??= SOMDoc.FromObject(payload);
        ListType ??= SOMDoc.SOMDocType(payload.GetType());

        List<Fact> ret = new() {
            new DynamicListFact(organizer)
        {
            payload = payload.Select(e => (object) e).ToList(),
            indirect_payload = indirect_payload,
            ListType = ListType,
            ServerDefinition = _ServerDefinition,
        }};

        if (_ServerDefinition == null)
            (ret[0] as DynamicListFact).SendToMMT();

        return ret;
    }

    public override MMTFact MakeMMTDeclaration()
        => indirect_payload == null
        ? throw new NotSupportedException(nameof(indirect_payload) + " must not be null")
        : new MMTGeneralFact(Label, ListType, indirect_payload);

    public static new List<Fact> parseFact(MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol)
            return new();

        List<Fact> orig = MMTFactory(null,
            MMTSymbol.defines, MMTSymbol.type,
            StageStatic.stage.factState, MMTSymbol.@ref
        );

        List<Fact> ret = new(orig);
        foreach (Fact Fact in orig)
            if (Fact is DynamicListFact pack)
                ret.AddRange(pack.UnpackMe(true)); //TODO: set to false, once Server can index

        return ret;
    }

    public SOMDoc SOMDocIndexer(int i)
        => new OMA(
            new OMS(MMTConstants.IndexList),
            new[] {
                indirect_payload,
                new OMLIT<float>(i)
        });

    // Best I can do is Equalness
    protected override bool EquivalentWrapped(DynamicListFact f1, DynamicListFact f2)
        => f1.payload.Count == f2.payload.Count
        && f1.ListType.Equivalent(f2.ListType)
        && f1.payload.Zip(f2.payload, (e1, e2) => e1.Equals(e2)).All(b => b);

    protected override string[] GetDependentFactIds()
        => indirect_payload.GetDependentFactIds();

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => new DynamicListFact(payload, indirect_payload?.MapURIs(old_to_new), ListType, organizer);

    public List<Fact> UnpackMe(bool BypassServer)
        => payload.SelectMany((object p, int i) =>
                MakeFact(p, SOMDocIndexer(i), _Facts, BypassServer)
            ).ToList();
}

/// <summary>
/// 
/// </summary>
/// <seealso cref="SOMDoc.MakeTupel(SOMDoc[])"/>
/// <remarks>Needed to refere to tupels serverside</remarks>
public class TupleFact : FactWrappedCRTP<TupleFact>
{
    public string[] lids = new string[0];

    public SOMDoc[] payload = new SOMDoc[0];


    public TupleFact() : base() { }

    public TupleFact(string[] lids, SOMDoc[] payload, FactRecorder organizer) : base(organizer)
    {
        Init(lids, payload);
        SendToMMT();
    }

    public TupleFact(string[] lids, SOMDoc[] payload, SOMDoc _ServerDefinition, FactRecorder organizer) : base(organizer)
    {
        Init(lids, payload);
        this.ServerDefinition = _ServerDefinition;
    }

    private void Init(string[] lids, SOMDoc[] payload)
    {
        this.lids = lids ?? new string[0];
        if (!HasDependentFacts)
        {
            if (payload.Any(p => p == null))
                throw new ArgumentException(nameof(payload) + "must not include null elements iff " + nameof(lids) + " is all nulls or empty!");

            this.payload = payload;
        }
        else
        {
            this.payload = new SOMDoc[Unity.Mathematics.math.max(lids.Length, payload.Length)];

            for (int i = 0; i < this.payload.Length; i++)
                if (lids[i] != null)
                    payload[i] = new OMS(lids[i]);
                else
                if (lids[i] == null && payload[i] == null)
                    throw new ArgumentException(nameof(lids) + " and " + nameof(payload) + " have to be complementary-not-null!");
        }
    }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => new TupleFact(
            lids.Select(id => id == null ? null : old_to_new[id]).ToArray(),
            payload.Select(somdoc => somdoc == null ? null : somdoc.MapURIs(old_to_new)).ToArray(),
            organizer);

    public static new TupleFact parseFact(MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol
         || MMTSymbol.defines is not OMA defineOMA)
        {
            return null;
        }

        string[] lids = defineOMA.arguments
            .Select(oms => (oms as OMS)?.uri)
            .ToArray();

        return new(lids, defineOMA.arguments, fact.@ref, StageStatic.stage.factState);
    }

    public override MMTFact MakeMMTDeclaration()
    {
        OMA Tupel = SOMDoc.MakeTupel(payload);
        return new MMTGeneralFact(Label, Tupel.SOMDocType(), Tupel);
    }

    protected override string[] GetDependentFactIds()
        => lids.Where(lid => lid != null).ToArray();

    protected override void RecalculateTransform() { }

    protected override bool EquivalentWrapped(TupleFact f1, TupleFact f2)
        => f1.payload.Length == f2.payload.Length
        && DependentFactsEquivalent(f1, f2)
        && f1.payload
            .Zip(f2.payload, (p1, p2) => (p1 is OMS && p2 is OMS) || SOMDoc.Equivalent(p1, p2))
            .All(b => b);
}

public class DynamicTupleFact : FactWrappedCRTP<DynamicTupleFact>, IUnpackable
{
    object payload;

    int count;

    [JsonIgnore]
    public Type[] Signature => _Signature ??= payload.GetType().GetGenericArguments();
    private Type[] _Signature;

    public SOMDoc indirect_payload;

    public SOMDoc TupleType;

    public DynamicTupleFact() : base() { }

    private DynamicTupleFact(FactRecorder organizer) : base(organizer) { }

    public static List<Fact> MMTFactory(object payload, SOMDoc indirect_payload, SOMDoc TupleType, FactRecorder organizer, SOMDoc _ServerDefinition = null)
    {
        if (payload == null
         && indirect_payload == null)
            return new();

        if (payload == null)
        {
            Func<object[], object[]> builder =
                indirect_payload.PartialInvokeCastingLambdaExpression(out _, out Type[] signature);

            if (signature.Length != 1)
                return new();

            payload = builder(new object[0])[0];
        }

        if (!TupleExtensions.IsTupleType(payload.GetType(), out int count))
            return new();
        Type[] Signature = payload.GetType().GetGenericArguments();
        List<Fact> ret = new();

        // Isomorphismen
        if (payload is Tuple<float, float, float> point)
        {
            Vector3 vec = new(point.Item1, point.Item2, point.Item3);
            ret.Add(new PointFact(vec, vec, organizer));
            return ret;
        }
        //if(payload is Tuple<Tuple<float, float>, Func<float, TIn>, Func<TIn?[..],TOut>> func)
        if (Signature.Length == 3
         && Signature[0].Equals(typeof(Tuple<float, float>))
         && FuncExtensions.IsFuncType(Signature[1], out int callerCount)
         && FuncExtensions.IsFuncType(Signature[2], out int calledCount)
         && callerCount == 2 && calledCount >= 2
         && Signature[1].GetGenericArguments()[0].Equals(typeof(float)))
        {
            Type callerType = Signature[1].GetGenericArguments()[1];
            Type[] calledType = Signature[2].GetGenericArguments();

            if ((calledCount == 2
                && callerType.Equals(calledType[0])
             ) || (callerType.IsGenericType
                && calledCount - 1 == callerType.GetGenericArguments().Length
                && callerType.GetGenericArguments()
                    .Zip(calledType[..^1], (a, b) => a.Equals(b))
                    .All(b => b)
            ))
            {
                Tuple<float, float> domain = (Tuple<float, float>)((ITuple)payload)[0];
                ret.Add(new FunctionFact(SOMDocIndexer(1, 3, indirect_payload), organizer));
                ret.Add(new FunctionFact(SOMDocIndexer(2, 3, indirect_payload), organizer));
                ret.Add(new FunctionCallFact(ret[1].Id, ret[0].Id, (domain.Item1, domain.Item2), organizer));
                return ret;
            }
        }

        return new(); // Don't care about Tuple without semantic meaning

        ////TODO? indirect_payload ??= SOMDoc.FromObject(payload);
        //TupleType ??= SOMDoc.SOMDocType(payload.GetType());

        //ret.Add(new DynamicTupleFact(organizer)
        //{
        //    payload = payload,
        //    indirect_payload = indirect_payload,
        //    TupleType = TupleType,
        //    count = count,
        //    ServerDefinition = _ServerDefinition,
        //});

        //if (_ServerDefinition == null)
        //    (ret[0] as DynamicTupleFact).SendToMMT();

        //return ret;
    }

    public override MMTFact MakeMMTDeclaration()
        => indirect_payload == null
        ? throw new NotSupportedException(nameof(indirect_payload) + " must not be null")
        : new MMTGeneralFact(Label, TupleType, indirect_payload);

    public static new List<Fact> parseFact(MMTFact fact)
    {
        if (fact is not MMTGeneralFact MMTSymbol)
            return new();

        return MMTFactory(null,
            MMTSymbol.defines, MMTSymbol.type,
            StageStatic.stage.factState, MMTSymbol.@ref
        );
    }

    public SOMDoc SOMDocIndexer(int i)
        => SOMDocIndexer(i, count, indirect_payload);

    public static SOMDoc SOMDocIndexer(int i, int count, SOMDoc indirect_payload)
    {
        if (i >= count && i < 0)
            throw new ArgumentOutOfRangeException($"{nameof(i)} must not be negative or exceed {count}");

        SOMDoc ret = indirect_payload;

        for (int j = 0; j < i; j++)
            ret = new OMA(
                new OMS(MMTConstants.ProjR),
                new[] { ret }
            );

        return count - 1 - i > 0
            ? new OMA(
                new OMS(MMTConstants.ProjL),
                new[] { ret })
            : ret;
    }

    // Best I can do is Equalness
    protected override bool EquivalentWrapped(DynamicTupleFact f1, DynamicTupleFact f2)
    {
        if (f1.count != f2.count
        || !f1.TupleType.Equivalent(f2.TupleType))
            return false;

        for (int i = 0; i < f1.count; i++)
            if (!((ITuple)f1.payload)[i].Equals(((ITuple)f2.payload)[i]))
                return false;

        return true;
    }

    protected override string[] GetDependentFactIds()
        => indirect_payload.GetDependentFactIds();

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => MMTFactory(payload, indirect_payload?.MapURIs(old_to_new), TupleType, organizer)?[0];

    public List<Fact> UnpackMe(bool BypassServer)
    {
        List<Fact> ret = new List<Fact>();

        for (int i = 0; i < count; i++)
            ret.AddRange(MakeFact(((ITuple)payload)[i], SOMDocIndexer(i), _Facts, BypassServer));

        return ret;
    }
}

public class RealLitFact : FactWrappedCRTP<RealLitFact>
{
    public float value;

    public RealLitFact() : base() { }

    public RealLitFact(float value, FactRecorder organizer) : base(organizer)
    {
        this.value = value;
        SendToMMT();
    }

    public RealLitFact(float value, SOMDoc _ServerDefinition, FactRecorder organizer) : base(organizer)
    {
        this.value = value;
        this.ServerDefinition = _ServerDefinition;
    }

    public override MMTFact MakeMMTDeclaration()
    {
        OMLIT<float> df = new(value);

        return new MMTGeneralFact(
                Label,
                df.SOMDocType(),
                df
            );
    }

    protected override bool EquivalentWrapped(RealLitFact f1, RealLitFact f2)
        => Mathf.Approximately(f1.value, f2.value);

    protected override string[] GetDependentFactIds()
        => new string[0];

    protected override void RecalculateTransform() { }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new, FactRecorder organizer)
        => new RealLitFact(value, organizer);
}